@@ -17,7 +17,7 @@
 %                                 March 2011                                  %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2019 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999-2020 ImageMagick Studio LLC, a non-profit organization      %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -81,7 +81,7 @@
 */
 #if defined(MAGICKCORE_WEBP_DELEGATE)
 static MagickBooleanType
-  WriteWEBPImage(const ImageInfo *,Image *,ExceptionInfo *);
+  WriteWEBPImage(const ImageInfo *,Image *);
 #endif
 
 /*
@@ -177,7 +177,7 @@ static MagickBooleanType IsWEBPImageLossless(const unsigned char *stream,
 #define CHUNK_HEADER_SIZE  8
 #define MAX_CHUNK_PAYLOAD  (~0U-CHUNK_HEADER_SIZE-1)
 
-  ssize_t
+  size_t
     offset;
 
   /*
@@ -191,7 +191,7 @@ static MagickBooleanType IsWEBPImageLossless(const unsigned char *stream,
     Read extended header.
   */
   offset=RIFF_HEADER_SIZE+TAG_SIZE+CHUNK_SIZE_BYTES+VP8X_CHUNK_SIZE;
-  while (offset+TAG_SIZE <= (ssize_t) (length-TAG_SIZE))
+  while (offset <= (length-TAG_SIZE-TAG_SIZE-4))
   {
     uint32_t
       chunk_size,
@@ -221,18 +221,19 @@ static int FillBasicWEBPInfo(Image *image,const uint8_t *stream,size_t length,
   webp_status=WebPGetFeatures(stream,length,features);
 
   if (webp_status != VP8_STATUS_OK)
-	return webp_status;
+    return(webp_status);
 
   image->columns=(size_t) features->width;
   image->rows=(size_t) features->height;
   image->depth=8;
   image->matte=features->has_alpha != 0 ? MagickTrue : MagickFalse;
 
-  return webp_status;
+  return(webp_status);
 }
 
 static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
-  size_t length,WebPDecoderConfig *configure,ExceptionInfo *exception)
+  size_t length,WebPDecoderConfig *configure,ExceptionInfo *exception,
+  MagickBooleanType is_first)
 {
   int
     webp_status;
@@ -240,7 +241,15 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
   register unsigned char
     *p;
 
+  size_t
+    canvas_width,
+    canvas_height,
+    image_width,
+    image_height;
+
   ssize_t
+    x_offset,
+    y_offset,
     y;
 
   WebPDecBuffer
@@ -249,16 +258,38 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
   MagickBooleanType
     status;
 
-  webp_status = FillBasicWEBPInfo(image,stream,length,configure);
-  if(webp_status != VP8_STATUS_OK)
-    return webp_status;
+  if (is_first)
+    {
+      canvas_width=image->columns;
+      canvas_height=image->rows;
+      x_offset=image->page.x;
+      y_offset=image->page.y;
+      image->page.x=0;
+      image->page.y=0;
+    }
+  else
+    {
+      x_offset=0;
+      y_offset=0;
+    }
+  webp_status=FillBasicWEBPInfo(image,stream,length,configure);
+  image_width=image->columns;
+  image_height=image->rows;
+  if (is_first)
+    {
+      image->columns=canvas_width;
+      image->rows=canvas_height;
+    }
+
+  if (webp_status != VP8_STATUS_OK)
+    return(webp_status);
 
   if (IsWEBPImageLossless(stream,length) != MagickFalse)
     image->quality=100;
 
-  webp_status=WebPDecode(stream,length, configure);
-  if(webp_status != VP8_STATUS_OK)
-    return webp_status;
+  webp_status=WebPDecode(stream,length,configure);
+  if (webp_status != VP8_STATUS_OK)
+    return(webp_status);
 
   p=(unsigned char *) webp_image->u.RGBA.rgba;
   for (y=0; y < (ssize_t) image->rows; y++)
@@ -274,10 +305,21 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
       break;
     for (x=0; x < (ssize_t) image->columns; x++)
     {
-      SetPixelRed(q,ScaleCharToQuantum(*p++));
-      SetPixelGreen(q,ScaleCharToQuantum(*p++));
-      SetPixelBlue(q,ScaleCharToQuantum(*p++));
-      SetPixelAlpha(q,ScaleCharToQuantum(*p++));
+      if ((x >= x_offset && x < x_offset + image_width) &&
+          (y >= y_offset && y < y_offset + image_height))
+        {
+          SetPixelRed(q,ScaleCharToQuantum(*p++));
+          SetPixelGreen(q,ScaleCharToQuantum(*p++));
+          SetPixelBlue(q,ScaleCharToQuantum(*p++));
+          SetPixelAlpha(q,ScaleCharToQuantum(*p++));
+        }
+      else
+        {
+          SetPixelRed(q,0);
+          SetPixelGreen(q,0);
+          SetPixelBlue(q,0);
+          SetPixelAlpha(q,0);
+        }
       q++;
     }
     if (SyncAuthenticPixels(image,exception) == MagickFalse)
@@ -298,7 +340,7 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
 
     WebPData
      chunk,
-     content = { stream, length };
+     content;
 
     WebPMux
       *mux;
@@ -306,6 +348,8 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
     /*
       Extract any profiles.
     */
+    content.bytes=stream;
+    content.size=length;
     mux=WebPMuxCreate(&content,0);
     (void) memset(&chunk,0,sizeof(chunk));
     WebPMuxGetFeatures(mux,&webp_flags);
@@ -342,7 +386,7 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
     WebPMuxDelete(mux);
   }
 #endif
-  return webp_status;
+  return(webp_status);
 }
 
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
@@ -350,51 +394,72 @@ static int ReadAnimatedWEBPImage(const ImageInfo *image_info,Image *image,
   uint8_t *stream,size_t length,WebPDecoderConfig *configure,
   ExceptionInfo *exception)
 {
-  WebPData data = {.bytes=stream, .size=length};
-
-  WebPDemuxer* demux = WebPDemux(&data);
-
-  WebPIterator iter;
+  Image
+    *original_image;
 
   int
-    webp_status = 0;
+    image_count,
+    webp_status;
 
-  int image_count = 0;
+  size_t
+    canvas_width,
+    canvas_height;
 
-  Image *original_image = image;
+  WebPData
+    data;
 
-  if (WebPDemuxGetFrame(demux, 1, &iter)) {
+  WebPDemuxer
+    *demux;
+
+  WebPIterator
+    iter;
+
+  image_count=0;
+  webp_status=0;
+  original_image=image;
+  webp_status=FillBasicWEBPInfo(image,stream,length,configure);
+  canvas_width=image->columns;
+  canvas_height=image->rows;
+  data.bytes=stream;
+  data.size=length;
+  demux=WebPDemux(&data);
+  if (WebPDemuxGetFrame(demux,1,&iter)) {
     do {
       if (image_count != 0)
         {
-	  AcquireNextImage(image_info,image);
-	  if (GetNextImageInList(image) == (Image *) NULL)
-	    {
-	      break;
-	    }
-	  image=SyncNextImageInList(image);
-	  CloneImageProperties(image, original_image);
-	  image->page.x = iter.x_offset;
-	  image->page.y = iter.y_offset;
+          AcquireNextImage(image_info,image);
+          if (GetNextImageInList(image) == (Image *) NULL)
+            break;
+          image=SyncNextImageInList(image);
+          CloneImageProperties(image, original_image);
+          image->page.x=iter.x_offset;
+          image->page.y=iter.y_offset;
+          webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
+            iter.fragment.size,configure,exception,MagickFalse);
         }
+      else
+        {
+          image->page.x=iter.x_offset;
+          image->page.y=iter.y_offset;
+          webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
+            iter.fragment.size,configure,exception,MagickTrue);
+        }
+      if (webp_status != VP8_STATUS_OK)
+        break;
 
-      webp_status = ReadSingleWEBPImage(image, iter.fragment.bytes,
-				        iter.fragment.size,
-					configure, exception);
-      if(webp_status != VP8_STATUS_OK)
-	break;
-
-      image->ticks_per_second = 100;
-      image->delay = iter.duration / 10;
-      if (image_info->verbose)
-	fprintf(stderr, "Reading WebP frame with delay %u\n", iter.duration);
+      image->page.width=canvas_width;
+      image->page.height=canvas_height;
+      image->ticks_per_second=100;
+      image->delay=iter.duration/10;
+      image->dispose=NoneDispose;
+      if (iter.dispose_method == WEBP_MUX_DISPOSE_BACKGROUND)
+        image->dispose=BackgroundDispose;
       image_count++;
-
     } while (WebPDemuxNextFrame(&iter));
     WebPDemuxReleaseIterator(&iter);
   }
   WebPDemuxDelete(demux);
-  return webp_status;
+  return(webp_status);
 }
 #endif
 
@@ -477,23 +542,15 @@ static Image *ReadWEBPImage(const ImageInfo *image_info,
 
   webp_status=FillBasicWEBPInfo(image,stream,length,&configure);
   if (webp_status == VP8_STATUS_OK) {
-    if (image_info->ping != MagickFalse)
-      {
-        stream=(unsigned char*) RelinquishMagickMemory(stream);
-        (void) CloseBlob(image);
-        return(GetFirstImageInList(image));
-      }
-
     if (configure.input.has_animation) {
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
-      webp_status=ReadAnimatedWEBPImage(image_info, image, stream,
-				 length, &configure, exception);
+      webp_status=ReadAnimatedWEBPImage(image_info,image,stream,length,
+        &configure,exception);
 #else
       webp_status=VP8_STATUS_UNSUPPORTED_FEATURE;
 #endif
     } else {
-      webp_status=ReadSingleWEBPImage(image, stream,
-		      length, &configure, exception);
+      webp_status=ReadSingleWEBPImage(image,stream,length,&configure,exception,MagickFalse);
     }
   }
 
@@ -589,7 +646,9 @@ ModuleExport size_t RegisterWEBPImage(void)
   entry->description=ConstantString("WebP Image Format");
   entry->mime_type=ConstantString("image/webp");
   entry->seekable_stream=MagickTrue;
+#if !defined(MAGICKCORE_WEBPMUX_DELEGATE)
   entry->adjoin=MagickFalse;
+#endif
   entry->magick_module=ConstantString("WEBP");
   entry->magick=(IsImageFormatHandler *) IsWEBP;
   if (*version != '\0')
@@ -683,17 +742,6 @@ typedef struct PictureMemory {
   struct PictureMemory *next;
 } PictureMemory;
 
-static void FreePictureMemoryList (PictureMemory* head) {
-  PictureMemory* next;
-  while(head != NULL) {
-    next = head->next;
-    if(head->pixel_info != NULL)
-      RelinquishVirtualMemory(head->pixel_info);
-    free(head);
-    head = next;
-  }
-}
-
 static MagickBooleanType WriteSingleWEBPImage(const ImageInfo *image_info,
   Image *image,WebPPicture *picture,PictureMemory *picture_memory,
   ExceptionInfo *exception)
@@ -759,36 +807,58 @@ static MagickBooleanType WriteSingleWEBPImage(const ImageInfo *image_info,
 }
 
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
+static void FreePictureMemoryList (PictureMemory* head) {
+  PictureMemory* next;
+  while(head != NULL) {
+    next = head->next;
+    if(head->pixel_info != NULL)
+      RelinquishVirtualMemory(head->pixel_info);
+    free(head);
+    head = next;
+  }
+}
+
 static MagickBooleanType WriteAnimatedWEBPImage(const ImageInfo *image_info,
-  Image *image, WebPConfig *configure,WebPMemoryWriter *writer_info,
+  Image *image,WebPConfig *configure,WebPMemoryWriter *writer_info,
   ExceptionInfo *exception)
 {
-  WebPAnimEncoderOptions
-	  enc_options;
-  WebPPicture
-	  picture;
-  WebPAnimEncoder
-	  *enc;
   Image
-	  *first_image;
-  PictureMemory
-	  *head, *current;
+    *first_image;
 
-  size_t frame_timestamp = 0,
-    effective_delta = 0;
+  PictureMemory
+    *current,
+    *head;
+
+  size_t
+    effective_delta = 0,
+    frame_timestamp = 0;
+
+  WebPAnimEncoder
+    *enc;
+
+  WebPAnimEncoderOptions
+    enc_options;
+
+  WebPData
+    webp_data;
+
+  WebPPicture
+    picture;
 
   WebPAnimEncoderOptionsInit(&enc_options);
   if (image_info->verbose)
     enc_options.verbose = 1;
 
-  image = CoalesceImages(image, exception);
-  first_image = image;
-  enc = WebPAnimEncoderNew(image->page.width, image->page.height, &enc_options);
+  image=CoalesceImages(image, exception);
+  first_image=image;
+  enc=WebPAnimEncoderNew((int) image->page.width,(int) image->page.height,
+    &enc_options);
 
-  head = calloc(sizeof(*head), 1);
-  current = head;
+  head=(PictureMemory *) calloc(sizeof(*head),1);
+  current=head;
 
-  while(image != NULL) {
+  while (image != NULL)
+  {
     if (WebPPictureInit(&picture) == 0)
       ThrowWriterException(ResourceLimitError,"UnableToEncodeImageFile");
 
@@ -796,21 +866,17 @@ static MagickBooleanType WriteAnimatedWEBPImage(const ImageInfo *image_info,
 
     effective_delta = image->delay*1000/image->ticks_per_second;
     if (effective_delta < 10)
-	    effective_delta = 100; // Consistent with gif2webp
+      effective_delta = 100; /* Consistent with gif2webp */
     frame_timestamp+=effective_delta;
 
-    if (image_info->verbose)
-	fprintf(stderr, "Writing WebP frame with delay %zu\n", effective_delta);
-
-    WebPAnimEncoderAdd(enc, &picture, frame_timestamp, configure);
+    WebPAnimEncoderAdd(enc,&picture,(int) frame_timestamp,configure);
 
     image = GetNextImageInList(image);
-    current->next = calloc(sizeof(*head), 1);
+    current->next=(PictureMemory *) calloc(sizeof(*head), 1);
     current = current->next;
   }
-  WebPData
-    webp_data = { writer_info->mem, writer_info->size };
-
+  webp_data.bytes=writer_info->mem;
+  webp_data.size=writer_info->size;
   WebPAnimEncoderAssemble(enc, &webp_data);
   WebPMemoryWriterClear(writer_info);
   writer_info->size=webp_data.size;
@@ -818,12 +884,12 @@ static MagickBooleanType WriteAnimatedWEBPImage(const ImageInfo *image_info,
   WebPAnimEncoderDelete(enc);
   DestroyImageList(first_image);
   FreePictureMemoryList(head);
-  return MagickTrue;
+  return(MagickTrue);
 }
 #endif
 
 static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
-  Image *image,ExceptionInfo * exception)
+  Image *image)
 {
   const char
     *value;
@@ -862,7 +928,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   if ((image->columns > 16383UL) || (image->rows > 16383UL))
     ThrowWriterException(ImageError,"WidthOrHeightExceedsLimit");
-  status=OpenBlob(image_info,image,WriteBinaryBlobMode,exception);
+  status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == MagickFalse)
     return(status);
   if (WebPConfigInit(&configure) == 0)
@@ -973,16 +1039,14 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
   if (WebPValidateConfig(&configure) == 0)
     ThrowWriterException(ResourceLimitError,"UnableToEncodeImageFile");
 
-  WriteSingleWEBPImage(image_info, image, &picture,
-    &memory, exception);
+  WriteSingleWEBPImage(image_info,image,&picture,&memory,&image->exception);
 
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
-  if ((GetPreviousImageInList(image) == (Image *) NULL) &&
+  if ((image_info->adjoin != MagickFalse) &&
+      (GetPreviousImageInList(image) == (Image *) NULL) &&
       (GetNextImageInList(image) != (Image *) NULL) &&
-      (image->iterations != 1)) {
-     WriteAnimatedWEBPImage(image_info, image, &configure,
-       &writer_info, exception);
-  }
+      (image->iterations != 1))
+    WriteAnimatedWEBPImage(image_info,image,&configure,&writer_info,&image->exception);
 #endif
 
   webp_status=WebPEncode(&configure,&picture);
@@ -1051,7 +1115,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
           break;
         }
       }
-      (void) ThrowMagickException(exception,GetMagickModule(),CorruptImageError,
+      (void) ThrowMagickException(&image->exception,GetMagickModule(),CorruptImageError,
         (char *) message,"`%s'",image->filename);
     }
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
@@ -1061,7 +1125,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
 
     WebPData
       chunk,
-      image_chunk = { writer_info.mem, writer_info.size };
+      image_chunk;
 
     WebPMux
       *mux;
@@ -1072,6 +1136,8 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
     /*
       Set image profiles (if any).
     */
+    image_chunk.bytes=writer_info.mem;
+    image_chunk.size=writer_info.size;
     mux_error=WEBP_MUX_OK;
     (void) memset(&chunk,0,sizeof(chunk));
     mux=WebPMuxNew();
@@ -1097,16 +1163,18 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
         mux_error=WebPMuxSetChunk(mux,"XMP",&chunk,0);
       }
     if (mux_error != WEBP_MUX_OK)
-      (void) ThrowMagickException(exception,GetMagickModule(),
+      (void) ThrowMagickException(&image->exception,GetMagickModule(),
         ResourceLimitError,"UnableToEncodeImageFile","`%s'",image->filename);
     if (chunk.size != 0)
       {
         WebPData
-          picture_profiles = { writer_info.mem, writer_info.size };
+          picture_profiles;
 
         /*
           Replace original container with image profile (if any).
         */
+        picture_profiles.bytes=writer_info.mem;
+        picture_profiles.size=writer_info.size;
         WebPMuxSetImage(mux,&image_chunk,1);
         mux_error=WebPMuxAssemble(mux,&picture_profiles);
         WebPMemoryWriterClear(&writer_info);
